import asyncio
import collections
import json
import threading
import urllib
import io

import PIL.Image

import tornado
import tornado.websocket

from .rpc_session import RPCSession

class RPCCallback(object):
    def __init__(self, object_type = None):
        self.ev = threading.Event()
        self.object_type = object_type

    def __call__(self, value):
        obj, bufs = value

        self.obj = obj
        if (self.object_type is not None):
            self.obj = self.object_type(self.obj)

        self.bufs = bufs

        self.ev.set()

    def wait(self):
        self.ev.wait()

class ClientConnection(object):
    def __init__(self, loop = None):
        if (loop is None):
            loop = asyncio.new_event_loop()

        self.loop = loop

        self.callbacks = {}

        self.next_id = 0

        self.rpc_session = RPCSession()
        self.vid_session = None


        th = threading.Thread(target = self.loop.run_forever)
        th.start()

    def __del__(self):
        self.close()

    async def _connect(self, host, port):
        url = f'ws://{host}:{port}'
        self.conn = await tornado.websocket.websocket_connect(url)

    def connect(self, address, port):
        fut = asyncio.run_coroutine_threadsafe(self._connect(address, port), self.loop)
        fut.result()

        fut = asyncio.run_coroutine_threadsafe(self.run(), self.loop)
        fut.add_done_callback(run_hook)

    def close(self):
        self.loop.call_soon_threadsafe(self.conn.close)
        self.loop.stop()

    @property
    def hostname(self):
        parts = urllib.parse.urlsplit(self.conn.request.url)
        return parts.hostname

    @property
    def port(self):
        parts = urllib.parse.urlsplit(self.conn.request.url)
        return parts.port

    def on(self, event, callback):
        try:
            callbacks = self.callbacks[event]
        except KeyError:
            callbacks = []
            self.callbacks[event] = callbacks

        callbacks.append(callback)

    def off(self, event, callback):
        callbacks = self.callbacks[event]
        callbacks.remove(callback)

    async def run(self):
        while True:
            msg = await self.conn.read_message()
            if (msg is None):
                if ('close' in self.callbacks):
                    for callback in self.callbacks['close']:
                        callback()
                return

            if isinstance(msg, bytes):
                if (self.vid_session is not None):
                    self.vid_session.add(msg)
            else:
                msg = json.loads(msg)
                buffer_count = msg.get('buffers', 0)

                bufs = []
                for buffer_index in range(buffer_count):
                    buf = await self.conn.read_message()
                    bufs.append(buf)

                event = msg['type']
                if (event == 'remote-call'):
                    self.rpc_session.set(msg['detail'], bufs)
                elif (event in self.callbacks):
                    type, sep, id = event.partition(':')
                    if (type == 'image'):
                        args = (PIL.Image.open(io.BytesIO(bufs[0])),)
                    elif (type == 'cancel'):
                        args = ()
                    else:
                        args = (msg['detail'], bufs)

                    callbacks = self.callbacks[event]
                    for callback in callbacks:
                        callback(*args)

    def new_id(self):
        value = self.next_id
        self.next_id += 1

        return value

    def remote_call(self, method, *args, object_type = None, wait = True, return_buffers = False):
        def func(callback, method, *args):
            id = self.rpc_session.get_id(callback)
            msg = {'detail': {'id': id, 'method': method, 'params': args}}

            self.conn.write_message(json.dumps(msg))

        on_result = RPCCallback(object_type)
        self.loop.call_soon_threadsafe(func, on_result, method, *args)

        if wait:
            on_result.wait()

            if return_buffers:
                return on_result.obj, on_result.bufs

            return on_result.obj

def run_hook(fut):
    e = fut.exception()
    if (e is not None):
        raise e



import datetime
import functools
import io
import numbers
import threading
import sys
import time

import PIL.Image

try:
    import IPython
except ImportError:
    pass

from .client_connection import ClientConnection
from .video import av_video

try:
    import ipydingo as ipdg
except ImportError:
    pass

try:
    import zeroconf
except ImportError:
    pass

def astimedelta(period):
    if isinstance(period, datetime.timedelta):
        return period

    if isinstance(period, numbers.Number):
        return datetime.timedelta(seconds = period)

    if isinstance(period, str):
        toks = period.split('.')

        t = datetime.datetime.strptime(toks[0], '%H:%M:%S')
        period = datetime.timedelta(hours = t.hour, minutes = t.minute, seconds = t.second)

        if (len(toks) == 2):
            t_us = datetime.datetime.strptime(toks[1], '%f')
            period += datetime.timedelta(microseconds = t_us.microsecond)

        return period

    raise Error('could not convert to timedelta')

def default_video_factory(hostname, port):
    if ('ipydingo' in sys.modules):
        shell = IPython.get_ipython()
        if ((shell is not None) and ('IPKernelApp' in shell.config)):
            return ipdg.ipy_video(hostname, port)

    return av_video(hostname, port)

class Client(object):
    def __init__(self, host = None, port = None):
        self.conn = None
        self.vid_conns = []

        if (host is None):
            self.autoconnect()
        else:
            if (port is None):
                port = 8080
            self.connect(host, port)

    @property
    def hostname(self):
        return self.conn.hostname

    @property
    def port(self):
        return self.conn.port

    def connect(self, address, port = 8080):
        self.conn = ClientConnection()
        self.conn.on('close', self.close)

        self.conn.connect(address, port)

    def autoconnect(self, type = '_http._tcp.local.'):
        listener = ZeroconfListener('dingo')

        zeroconf_cli = zeroconf.Zeroconf()
        zeroconf_cli.add_service_listener(type, listener)

        listener.wait()
        zeroconf_cli.close()

        self.connect(listener.host, listener.port)

    def close(self):
        while (len(self.vid_conns) > 0):
            vid_conn = self.vid_conns.pop()
            vid_conn.close()

        if (self.conn is not None):
            self.conn.close()
            self.conn = None

    def on(self, event, callback):
        self.conn.on(event, callback)

    def off(self, event, callback):
        self.conn.off(event, callback)

    def remove_listener(self, ev, callback):
        pass

    def camera_brightness(self):
        return self.conn.remote_call('camera_brightness')

    def camera_brightness_range(self):
        return self.conn.remote_call('camera_brightness_range', object_type = tuple)

    def camera_exposure(self):
        return self.conn.remote_call('camera_exposure')

    def camera_exposure_range(self):
        return self.conn.remote_call('camera_exposure_range', object_type = tuple)

    def image(self, callback = None, wait = True):
        def default_callback(im):
            nonlocal res
            res = im

        def cancel_callback():
            self.off(str.format('cancel:{}', id), cancel_callback)
            self.off(str.format('image:{}', id), callback)

            ev.set()

        if (callback is None):
            callback = default_callback

        id = self.conn.new_id()

        self.on(str.format('image:{}', id), callback)
        self.on(str.format('cancel:{}', id), cancel_callback)

        ev = threading.Event()
        res = None
        self.conn.remote_call('image', id)

        if wait:
            ev.wait()

        return res

    def image_every(self, frequency, expiry, callback = None, wait = True):
        @functools.wraps(callback)
        def wrapper(callback):
            def cancel_callback():
                self.off(str.format('cancel:{}', id), cancel_callback)
                self.off(str.format('image:{}', id), callback)

                ev.set()

            id = self.conn.new_id()

            self.on(str.format('image:{}', id), callback)
            self.on(str.format('cancel:{}', id), cancel_callback)

            ev = threading.Event()
            self.conn.remote_call('image_every', id, str(frequency), str(expiry))

            if wait:
                ev.wait()

            return callback

        frequency = astimedelta(frequency)
        expiry = astimedelta(expiry)

        if (callback is None):
            return wrapper

        wrapper(callback)

    def image_background(self):
        obj, bufs = self.conn.remote_call('image_background', return_buffers = True)
        if (len(bufs) == 0):
            return None

        return PIL.Image.open(io.BytesIO(bufs[0]))

    def render_state(self):
        return self.conn.remote_call('render_state')

    def render_z(self):
        return self.conn.remote_call('render_z')

    def state(self):
        return self.conn.remote_call('state')

    def reset_background(self):
        self.conn.remote_call('resetBackground')

    def image_selection(self):
        return self.conn.remote_call('image_selection', object_type = tuple)

    def image_size(self):
        return self.conn.remote_call('image_size')

    def image_size_max(self):
        return self.conn.remote_call('image_size_max')

    def set_camera_brightness(self, value):
        return self.conn.remote_call('set_camera_brightness', value)

    def set_camera_exposure(self, value):
        return self.conn.remote_call('set_camera_exposure', value)

    def set_background_policy(self, value):
        self.conn.remote_call('setBackgroundPolicy', value)

    def set_image_maximized(self):
        self.conn.remote_call('set_image_maximized')

    def set_image_selection(self, value):
        self.conn.remote_call('set_image_selection', value)

    def set_render_state(self, value):
        self.conn.remote_call('set_render_state', value)

    def set_render_remove_background(self, value):
        self.conn.remote_call('set_render_remove_background', value)

    def set_video_resolution(self, value):
        self.conn.remote_call('setVideoResolution', value)

    def save_state(self):
        self.conn.remote_call('save_state')

    def touch(self):
        return self.conn.remote_call('touch')

    def video(self, video_factory = default_video_factory):
        if (video_factory == 'av'):
            video_factory = av_video
        elif (video_factory == 'ipy'):
            video_factory = ipdg.ipy_video

        vid, vid_conn = video_factory(self.hostname, self.port)
        self.vid_conns.append(vid_conn)

        return vid

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.close()

class ZeroconfListener(object):
    def __init__(self, name):
        self.name = name
        self.host = None
        self.port = None
        self.ev = threading.Event()

    def add_service(self, zeroconf, type, name):
        info = zeroconf.get_service_info(type, name)

        if (info.get_name() == self.name):
            addresses =  info.parsed_addresses()

            self.host = addresses[0]
            self.port = info.port

            self.ev.set()

    def update_service(self, zeroconf, type, name):
        pass

    def remove_service(self, zeroconf, type, name):
        pass

    def wait(self):
        self.ev.wait()
